import heat as ht

e = ht.e
"""IEEE 754 floating-point representation of Euler's constant."""
inf = ht.inf
"""IEEE 754 floating-point representation of (positive) infinity."""
nan = ht.nan
"""IEEE 754 floating-point representation of Not a Number (``NaN``)."""
newaxis = None
"""An alias for ``None`` which is useful for indexing arrays."""
pi = ht.pi
"""IEEE 754 floating-point representation of the mathematical constant :math:`\\pi`."""
